# Code used for testing the consistency between Matlab and Python versions of the metric.

from utils import load_mat_tensor
import torch
import numpy as np
import os

class FovVideoVDP_Testbench():
    def __init__(self, matlab_intermediate_path = "D:\\work\\st_fov_metric\\matlab_intermediate\\"):
        self.tests_passed = 0
        self.tests_failed = 0
        self.matlab_intermediate_path = matlab_intermediate_path

    def print_summary(self):
        total  = self.tests_passed + self.tests_failed
        if total > 0:
            perc = 100.0 * float(self.tests_passed) / float(total)
            print("%d passed of %d total (%0.2f %%)" % (self.tests_passed, total, perc))

    def verify_against_matlab(self, x, tag, device, file = None, tolerance = 0.001, verbose=False, relative=False):
        if file is None:
            mat_tensor = load_mat_tensor(os.path.join(self.matlab_intermediate_path, "%s.mat" % tag), tag, device)
        else:
            mat_tensor = load_mat_tensor(os.path.join(self.matlab_intermediate_path, "%s.mat" % file), tag, device)

        if mat_tensor is None:
            print("Error: cannot find file for tag " + tag)
            return 

        mat_tensor = torch.squeeze(mat_tensor)
        x = torch.squeeze(x)

        assert mat_tensor.shape == x.shape, "Tensor '%s' shape (%s) does not match MATLAB (%s)" % (tag, str(x.shape), str(mat_tensor.shape))

        if relative:
            difftype = "relative"
            diff = torch.abs(x - mat_tensor) / (torch.abs(mat_tensor) + 0.00001)
        else:
            difftype = "absolute"
            diff = torch.abs(x - mat_tensor)

        max_diff = torch.max(diff).cpu().numpy()

        if max_diff > tolerance:
            self.tests_failed += 1
            if len(x.shape)>=1:
                # Tensor
                diff_loc = np.unravel_index([torch.argmax(torch.abs(x - mat_tensor)).cpu().numpy()], x.shape)
                diff_loc = tuple([d[0] for d in diff_loc])
                mean_diff = torch.mean(diff.type(torch.cuda.FloatTensor)).cpu().numpy()
                print("[FAIL] Max %s error for '%s' is %f (> %f) @ %s (mean %f, %f vs %f (matlab))" % (difftype, tag, max_diff, tolerance, str(diff_loc), mean_diff, x[diff_loc], mat_tensor[diff_loc]))
                if verbose:
                    if len(x.shape)==2:
                        for yd in range(-2,3):
                            for xd in range(-2,3):
                                if ((diff_loc[0] + yd) >= 0 and (diff_loc[0] + yd) < x.shape[0] and
                                    (diff_loc[1] + xd) >= 0 and (diff_loc[1] + xd) < x.shape[1]):
                                   print("%0.4f " % x[diff_loc[0] + yd, diff_loc[1] + xd], end='')
                            print('   ', end='')
                            for xd in range(-2,3):
                                if ((diff_loc[0] + yd) >= 0 and (diff_loc[0] + yd) < x.shape[0] and
                                    (diff_loc[1] + xd) >= 0 and (diff_loc[1] + xd) < x.shape[1]):
                                   print("%0.4f " % mat_tensor[diff_loc[0] + yd, diff_loc[1] + xd], end='')
                            print('')
            else:
                # Scalar
                print("[FAIL] Error for '%s' is %f (> %f), %f vs %f (matlab)" % (tag, max_diff, tolerance, x.cpu().item(), mat_tensor.cpu().item()))

            if verbose:
                print("    [Python] " + str(x.flatten()[0:4]))#[...,0:3,0:3])
                print("    [Matlab] " + str(mat_tensor.flatten()[0:4]))#[...,0:3,0:3])
        else:
            self.tests_passed += 1
            if len(x.shape)>=1:
                print("[PASS] Tensor %s matches MATLAB (tolerance %f, max_diff %f)" % (tag, tolerance, max_diff))
            else:
                print("[PASS] Tensor %s matches MATLAB (tolerance %f, max_diff %f,  %f vs %f (matlab))" % (tag, tolerance, max_diff, x.cpu().item(), mat_tensor.cpu().item()))



from pyfvvdp.fvvdp import fvvdp
from pyfvvdp.video_source_file import fvvdp_video_source_file, load_image_as_array
from pyfvvdp.fvvdp_display_model import fvvdp_display_photo_gog, fvvdp_display_photo_absolute, fvvdp_display_geometry




import os
from third_party.loadmat import loadmat
import torch
import numpy as np
import json
import torch.nn.functional as Func
from PIL import Image

def torch_gpu_mem_info():
    t = torch.cuda.get_device_properties(0).total_memory
    c = torch.cuda.memory_cached(0)
    a = torch.cuda.memory_allocated(0)
    f = c-a  # free inside cache
    print("GPU mem used: %d M (cache %d M)" % (a/(1024*1024), c/(1024*1024)))

def json2dict(file):
    data = None
    if os.path.isfile(file):
        with open(file, "r") as json_file:
            data=json.load(json_file)
    else:
        print("Error: Cannot find file %s" % file)
    return data

def linear2srgb_torch(lin):
    lin = torch.clamp(lin, 0.0, 1.0)
    srgb = torch.where(lin > 0.0031308, (1.055 * (lin ** (1/2.4))) - 0.055, 12.92 * lin)
    return srgb

def srgb2linear_torch(srgb):
    srgb = torch.clamp(srgb, 0.0, 1.0)
    lin = torch.where(srgb > 0.04045, ((srgb + 0.055) / 1.055)**2.4, srgb/12.92)
    return lin

def img2np(img):
    return np.array(img, dtype="float32") * 1.0/255.0

def np2img(nparr):
    return Image.fromarray(np.clip(nparr * 255.0, 0.0, 255.0).astype('uint8'))

def l2rgb(x):
    return np.concatenate([x,x,x], -1)

def stack_horizontal(nparr):
    return np.concatenate([nparr[i] for i in range(len(nparr))], axis=-2)

def stack_vertical(nparr):
    return np.concatenate([nparr[i] for i in range(len(nparr))], axis=-3)


def load_mat_dict(filepath, data_label, device):
    # datapath = "D:\\work\\st_fov_metric"
    # filepath = os.path.join(datapath, rel_path)

    if not os.path.isfile(filepath):
        return None
    else:
        v = loadmat(filepath)
        if data_label in v:
            return v[data_label]
        else:
            print("Cannot find key %s, valid keys are %s" % (data_label, v.keys()))
            return None

def load_mat_tensor(filepath, data_label, device):
    # datapath = "D:\\work\\st_fov_metric"
    # filepath = os.path.join(datapath, rel_path)

    if not os.path.isfile(filepath):
        return None
    else:
        v = loadmat(filepath)
        if data_label in v:
            return torch.tensor(v[data_label], device=device)
        else:
            print("Cannot find key %s, valid keys are %s" % (data_label, v.keys()))
            return None


# args are indexed at 0
def fovdots_load_ref(content_id, device, data_res="full"):
    if data_res != "full":
        print("Note: Using data resolution %s" % (data_res))
    hwd_tensor = load_mat_tensor("D:\\work\\st_fov_metric\\data_vid_%s\\content_%d_ref.mat" % (data_res, content_id+1), "I_vid", device)
    dhw_tensor = hwd_tensor.permute(2,0,1)
    ncdhw_tensor = torch.unsqueeze(torch.unsqueeze(dhw_tensor, 0), 0)
    return ncdhw_tensor

# args are indexed at 0
def fovdots_load_condition(content_id, condition_id, device, data_res="full"):
    if data_res != "full":
        print("Note: Using data resolution %s" % (data_res))
    hwd_tensor = load_mat_tensor("D:\\work\\st_fov_metric\\data_vid_%s\\content_%d_condition_%d.mat" % (data_res, content_id+1, condition_id+1), "I_vid", device)
    dhw_tensor = hwd_tensor.permute(2,0,1)
    ncdhw_tensor = torch.unsqueeze(torch.unsqueeze(dhw_tensor, 0), 0)
    return ncdhw_tensor


class ImGaussFilt():
    def __init__(self, sigma, device):
        self.filter_size = 2 * int(np.ceil(2.0 * sigma)) + 1
        self.half_filter_size = (self.filter_size - 1)//2

        self.K = torch.zeros((1, 1, self.filter_size, self.filter_size), device=device)

        for ii in range(self.filter_size):
            for jj in range(self.filter_size):
                distsqr = float(ii - self.half_filter_size) ** 2 + float(jj - self.half_filter_size) ** 2
                self.K[0,0,jj,ii] = np.exp(-distsqr / (2.0 * sigma * sigma))

        self.K = self.K/self.K.sum()

    def run(self, img):
        
        if len(img.shape) == 2: img_4d = img.reshape((1,1,img.shape[0],img.shape[1]))
        else:                   img_4d = img

        pad = (
            self.half_filter_size,
            self.half_filter_size,
            self.half_filter_size,
            self.half_filter_size,)

        img_4d = Func.pad(img_4d, pad, mode='reflect')
        return Func.conv2d(img_4d, self.K)[0,0]



import torch
import numpy as np 
import os
import sys
import math

# x_q : query tensor 
# x   : boundaries tensor
# inspired from: https://github.com/sbarratt/torch_interpolations/blob/master/torch_interpolations/multilinear.py#L39
def get_interpolants_v1(x_q, x):
    imax = torch.bucketize(x_q, x)
    imax[imax >= x.shape[0]] = x.shape[0] - 1
    imin = (imax - 1).clamp(0, x.shape[0] - 1)

    ifrc = (x_q - x[imin]) / (x[imax] - x[imin] + 0.000001)
    ifrc[imax == imin] = 0.
    ifrc[ifrc < 0.0] = 0.

    return imin, imax, ifrc

def get_interpolants_v0(x_q, x, device):
    imin = torch.zeros(x_q.shape, dtype=torch.long).to(device)
    ifrc = torch.zeros(x_q.shape, dtype=torch.float32).to(device)
    N = x.shape[0]
    for i in range(N):
        if i==0:
            imin  = torch.where(x_q  <= x[i], torch.tensor(i, dtype=torch.long).to(device),  imin)
            ifrc  = torch.where(x_q  <= x[i], torch.tensor(0.).to(device), ifrc)

        if i==(N-1):
            imin  = torch.where(x[i] <= x_q,  torch.tensor(i, dtype=torch.long).to(device),  imin)
            ifrc  = torch.where(x[i] <= x_q,  torch.tensor(0.).to(device), ifrc)
        else:
            t = (x_q - x[i])/(x[i+1] - x[i])
            imin  = torch.where((x[i] <= x_q) & (x_q < x[i+1]), torch.tensor(i,dtype=torch.long).to(device), imin)
            ifrc  = torch.where((x[i] <= x_q) & (x_q < x[i+1]), t, ifrc)

    imax = torch.min(imin+1, torch.tensor(N-1, dtype=torch.long).to(device))

    return imin, imax, ifrc

def interp3(x, y, z, v, x_q, y_q, z_q):
    shp = x_q.shape
    x_q = x_q.flatten()
    y_q = y_q.flatten()
    z_q = z_q.flatten()

    imin, imax, ifrc = get_interpolants_v1(x_q, x)
    jmin, jmax, jfrc = get_interpolants_v1(y_q, y)
    kmin, kmax, kfrc = get_interpolants_v1(z_q, z)

    filtered = (
        ((v[jmin,imin,kmin] * (1.0-ifrc) + v[jmin,imax,kmin] * (ifrc)) * (1.0-jfrc) + 
         (v[jmax,imin,kmin] * (1.0-ifrc) + v[jmax,imax,kmin] * (ifrc)) *     (jfrc)) * (1.0 - kfrc) + 
        ((v[jmin,imin,kmax] * (1.0-ifrc) + v[jmin,imax,kmax] * (ifrc)) *     (1.0-jfrc) + 
         (v[jmax,imin,kmax] * (1.0-ifrc) + v[jmax,imax,kmax] * (ifrc)) *     (jfrc)) * (kfrc))

    return filtered.reshape(shp)

def interp1(x, v, x_q):
    shp = x_q.shape
    x_q = x_q.flatten()

    imin, imax, ifrc = get_interpolants_v1(x_q, x)

    filtered = v[imin] * (1.0-ifrc) + v[imax] * (ifrc) 

    return filtered.reshape(shp)


def test_interp3(device):
    x_q = torch.tensor([0.5, 1.9, 2.1], dtype=torch.float32).to(device)
    y_q = torch.tensor([2.0, 2.0, 2.0], dtype=torch.float32).to(device)
    z_q = torch.tensor([1.5, 2.0, 2.0], dtype=torch.float32).to(device)

    x = torch.tensor([1.0, 2.0, 3.0], dtype=torch.float32).to(device)
    y = torch.tensor([1.0, 2.0, 3.0], dtype=torch.float32).to(device)
    z = torch.tensor([1.0, 2.0, 3.0], dtype=torch.float32).to(device)
    v = torch.tensor([
        [
        [10.0, 20.0, 30.0],
        [15.0, 30.0, 45.0],
        [20.0, 40.0, 60.0]],
        [
        [100.0, 200.0, 300.0],
        [150.0, 300.0, 450.0],
        [200.0, 400.0, 600.0]],
        [
        [1000.0, 2000.0, 3000.0],
        [1500.0, 3000.0, 4500.0],
        [2000.0, 4000.0, 6000.0]],
        ], dtype=torch.float32).to(device)

    print(x_q)
    print(x)
    print(v)
    print(interp3(x, y, z, v, x_q, y_q, z_q))


if __name__ == '__main__':
    test_interp3(torch.device('cpu'))



# Classes for reading images or videos from files so that they can be passed to FovVideoVDP frame-by-frame

import os
import imageio.v2 as io
import numpy as np
from torch.functional import Tensor
import torch
import ffmpeg

from pyfvvdp.fvvdp import fvvdp_video_source, fvvdp_video_source_dm, fvvdp_video_source_array, reshuffle_dims

try:
    # This may fail if OpenEXR is not installed. To install,
    # ubuntu: sudo apt install libopenexr-dev
    # mac: brew install openexr
    import pyexr
    use_pyexr = True
except ImportError as e:
    # Imageio's imread is unreliable for OpenEXR images
    # See https://github.com/imageio/imageio/issues/517
    use_pyexr = False

def load_image_as_array(imgfile):
    ext = os.path.splitext(imgfile)[1].lower()
    if ext == '.exr' and use_pyexr:
        precisions = pyexr.open(imgfile).precisions
        assert precisions.count(precisions[0]) == len(precisions), 'All channels must have same precision'
        img = pyexr.read(imgfile, precision=precisions[0])
    else:
        # 16-bit PNG not supported by default
        lib = 'PNG-FI' if ext == '.png' else None
        img = io.imread(imgfile, format=lib)
    return img


class video_reader:

    def __init__(self, vidfile, frames):
        try:
            probe = ffmpeg.probe(vidfile)
        except:
            raise RuntimeError("ffmpeg failed to open file \"" + vidfile + "\"")

        # select the first video stream
        video_stream = next((stream for stream in probe['streams'] if stream['codec_type'] == 'video'), None)

        self.width = int(video_stream['width'])
        self.height = int(video_stream['height'])
        num_frames = int(video_stream['nb_frames'])
        avg_fps_num, avg_fps_denom = [float(x) for x in video_stream['r_frame_rate'].split("/")]
        self.avg_fps = avg_fps_num/avg_fps_denom

        if frames==-1:
            self.frames = num_frames
        else:    
            if num_frames < frames:
                err_str = 'Expecting {needed_frames} frames but only {available_frames} available in the file \"{file}\"'.format( needed_frames=frames, available_frames=num_frames, file=vidfile)
                raise RuntimeError( err_str )
            self.frames = frames

        self.process = (
            ffmpeg
            .input(vidfile)
            .output('pipe:', format='rawvideo', pix_fmt='rgb24')
            .run_async(pipe_stdout=True, quiet=True)
        )

        self.curr_frame = -1

    def get_frame(self):
        in_bytes = self.process.stdout.read(self.width * self.height * 3)
        if not in_bytes or self.curr_frame == self.frames:
            return None
        in_frame = (
            np
            .frombuffer(in_bytes, np.uint8)
            .reshape([self.height, self.width, 3])
        ) 
        self.curr_frame += 1
        return in_frame       


'''
Use ffmpeg to read video frames, one by one.
'''
class fvvdp_video_source_video_file(fvvdp_video_source_dm):

    def __init__( self, test_fname, reference_fname, display_photometry='sdr_4k_30', color_space_name='sRGB', frames=-1 ):

        super().__init__(display_photometry=display_photometry, color_space_name=color_space_name)        

        self.test_vidr = video_reader(test_fname, frames)
        self.reference_vidr = video_reader(reference_fname, frames)
        if frames==-1:
            self.frames = self.test_vidr.frames
        else:
            self.frames = frames

        if self.test_vidr.height != self.reference_vidr.height or self.test_vidr.width != self.reference_vidr.width:
            raise RuntimeError( 'Test and reference video sequences must have the same resolutions. Found: test {tw}x{th}, reference {rw}x{rh}'.format(tw=self.test_vidr.width, th=self.reference_vidr.height, rw=self.test_vidr.width, rh=self.referemce_vidr.height) )

        # self.last_test_frame = None
        # self.last_reference_frame = None
        
    # Return (height, width, frames) touple with the resolution and
    # the length of the video clip.
    def get_video_size(self):
        return (self.test_vidr.height, self.test_vidr.width, self.frames )

    # Return the frame rate of the video
    def get_frames_per_second(self) -> int:
        return self.test_vidr.avg_fps
    
    # Get a pair of test and reference video frames as a single-precision luminance map
    # scaled in absolute inits of cd/m^2. 'frame' is the frame index,
    # starting from 0. 
    def get_test_frame( self, frame, device ) -> Tensor:
        # if not self.last_test_frame is None and frame == self.last_test_frame[0]:
        #     return self.last_test_frame[1]
        L = self._get_frame( self.test_vidr, frame, device )
        # self.last_test_frame = (frame,L)
        return L

    def get_reference_frame( self, frame, device ) -> Tensor:
        # if not self.last_reference_frame is None and frame == self.last_reference_frame[0]:
        #     return self.last_reference_frame[1]
        L = self._get_frame( self.reference_vidr, frame, device )
        # self.reference_test_frame = (frame,L)
        return L

    def _get_frame( self, vid_reader, frame, device ):        

        if frame != (vid_reader.curr_frame+1):
            raise RuntimeError( 'Video can be currently only read frame-by-frame. Random access not implemented.' )

        frame_np = vid_reader.get_frame()

        if frame_np is None:
            raise RuntimeError( 'Could not read frame {}'.format(frame) )

        frame_t_hwc = torch.tensor(frame_np)
        assert frame_t_hwc.dtype is torch.uint8
        frame_t = reshuffle_dims( frame_t_hwc, in_dims='HWC', out_dims="BCFHW" )
        frame_t = frame_t.to(device).to(torch.float32)/255

        L = self.dm_photometry.forward( frame_t )

        # Convert to grayscale
        L = L[:,0:1,:,:,:]*self.color_to_luminance[0] + L[:,1:2,:,:,:]*self.color_to_luminance[1] + L[:,2:3,:,:,:]*self.color_to_luminance[2]

        return L

'''
Recognize whether the file is an image of video and wraps an appropriate video_source for the given content.
'''
class fvvdp_video_source_file(fvvdp_video_source):

    def __init__( self, test_fname, reference_fname, display_photometry='sdr_4k_30', color_space_name='sRGB', frames=-1 ):
        # these extensions switch mode to images instead
        image_extensions = [".png", ".jpg", ".gif", ".bmp", ".jpeg", ".ppm", ".tiff", ".dds", ".exr", ".hdr"]

        assert os.path.isfile(test_fname), f'File does not exists: "{test_fname}"'
        assert os.path.isfile(reference_fname), f'File does not exists: "{reference_fname}"'

        if os.path.splitext(test_fname)[1].lower() in image_extensions:
            assert os.path.splitext(reference_fname)[1].lower() in image_extensions, 'Test is an image, but reference is a video'
            img_test = load_image_as_array(test_fname)
            img_reference = load_image_as_array(reference_fname)
            self.vs = fvvdp_video_source_array( img_test, img_reference, 0, dim_order='HWC', display_photometry=display_photometry, color_space_name=color_space_name )
        else:
            assert os.path.splitext(reference_fname)[1].lower() not in image_extensions, 'Test is a video, but reference is an image'
            self.vs = fvvdp_video_source_video_file( test_fname, reference_fname, display_photometry=display_photometry, color_space_name=color_space_name, frames=frames )


    # Return (height, width, frames) touple with the resolution and
    # the length of the video clip.
    def get_video_size(self):
        return self.vs.get_video_size()

    # Return the frame rate of the video
    def get_frames_per_second(self) -> int:
        return self.vs.get_frames_per_second()
    
    # Get a pair of test and reference video frames as a single-precision luminance map
    # scaled in absolute inits of cd/m^2. 'frame' is the frame index,
    # starting from 0. 
    def get_test_frame( self, frame, device ) -> Tensor:
        return self.vs.get_test_frame( frame, device )

    def get_reference_frame( self, frame, device ) -> Tensor:
        return self.vs.get_reference_frame( frame, device )



# Command-line interface for FovVideoVDP. 

import os, sys
import os.path
import argparse
import logging
#from natsort import natsorted
import glob
import ffmpeg
import numpy as np
import torch
import imageio.v2 as imageio

import pyfvvdp
# from pyfvvdp.visualize_diff_map import visualize_diff_map
#from pytorch_msssim import SSIM
from utils import *

def expand_wildcards(filestrs):
    if not isinstance(filestrs, list):
        return [ filestrs ]
    files = []
    for filestr in filestrs:
        if "*" in filestr:
            curlist = glob.glob(filestr)
            files = files + curlist
        else:
            files.append(filestr)
    return files

# Save a numpy array as a video
def np2vid(np_srgb, vidfile, fps, verbose=False):

    N, H, W, C = np_srgb.shape
    if C == 1:
        np_srgb = np.concatenate([np_srgb]*3, -1)
    process = (
        ffmpeg
            .input('pipe:', format='rawvideo', pix_fmt='rgb24', s='{}x{}'.format(W, H), r=fps)
            .output(vidfile, pix_fmt='yuv420p', crf=10)
            .overwrite_output()
            .global_args( '-hide_banner')
            .global_args( '-loglevel', 'info' if verbose else 'quiet')
            .run_async(pipe_stdin=True)
    )
    for fid in range(N):
        process.stdin.write(
                (np_srgb[fid,...] * 255.0)
                .astype(np.uint8)
                .tobytes()
        )
    process.stdin.close()
    process.wait()

# Save a numpy array as an image
def np2img(np_srgb, imgfile):

    N, H, W, C = np_srgb.shape
    if C == 1:
        np_srgb = np.concatenate([np_srgb]*3, -1)

    if N>1:
        sys.exit("Expecting an image, found video")

    imageio.imwrite( imgfile, (np.clip(np_srgb,0.0,1.0)[0,...]*255.0).astype(np.uint8) )

# -----------------------------------
# Command-line Arguments
# -----------------------------------
def parse_args():
    parser = argparse.ArgumentParser(description="Evaluate FovVideoVDP on a set of videos")
    parser.add_argument("--test", type=str, nargs='+', required = True, help="list of test images/videos")
    parser.add_argument("--ref", type=str, nargs='+', required = True, help="list of reference images/videos")
    parser.add_argument("--gpu", type=int,  default=-1, help="select which GPU to use (e.g. 0), default is CPU")
    parser.add_argument("--heatmap", type=str, default="none", help="type of difference map (none, raw, threshold, supra-threshold)")
    parser.add_argument("--heatmap-dir", type=str, default=None, help="in which directory heatmaps should be stored (the default is the current directory)")
    parser.add_argument("--verbose", action='store_true', default=False, help="Verbose mode")
    parser.add_argument("--foveated", action='store_true', default=False, help="Run in a foveated mode (non-foveated is the default)")
    parser.add_argument("--display", type=str, default="standard_4k", help="display name, e.g. HTC Vive")
    #parser.add_argument("--nframes", type=int, default=60, help="# of frames from video you want to load")
    parser.add_argument("--quiet", action='store_const', const=True, default=False, help="Do not print any information but the final JOD value.")
    args = parser.parse_args()
    return args

def main():
    args = parse_args()

    if args.quiet:
        log_level = logging.WARNING
    else:
        log_level = logging.INFO
        
    logging.basicConfig(format='[%(levelname)s] %(message)s', level=log_level)

    if args.gpu >= 0 and torch.cuda.is_available():
        device = torch.device('cuda:' + str(args.gpu))
    else:
        device = torch.device('cpu')

    logging.info("Running on device: " + str(device))

    # heatmap_types = {
    #     "threshold"   : {"scale" : 1.000, "colormap_type": "trichromatic"},
    #     "supra-threshold" : {"scale" : 0.333, "colormap_type": "dichromatic"},
    # }
    heatmap_types = ["raw", "threshold", "supra-threshold"]

    if args.heatmap == "none":
        args.heatmap = None

    if args.heatmap:
        if not args.heatmap in heatmap_types:
            logging.error( 'The recognized heatmap types are: "none", "raw", "threshold" and "supra-threshold"' )
            sys.exit()

        do_heatmap = True
    else:
        do_heatmap = False

    args.test = expand_wildcards(args.test)
    args.ref = expand_wildcards(args.ref)    

    N_test = len(args.test)
    N_ref = len(args.ref)

    if N_test==0:
        logging.error( "No test images/videos found." )
        sys.exit()

    if N_ref==0:
        logging.error( "No reference images/videos found." )
        sys.exit()

    if N_test != N_ref and N_test != 1 and N_ref != 1:
        logging.error( "Pass the same number of reference and test sources, or a single reference (to be used with all test sources), or a single test (to be used with all reference sources)." )
        sys.exit()

    fv = pyfvvdp.fvvdp( display_name=args.display, foveated=args.foveated, heatmap=args.heatmap, device=device )

    logging.info( 'When reporting metric results, please include the following information:' )    

    if args.display.startswith('standard_'):
        #append this if are using one of the standard displays
        standard_str = ', (' + args.display + ')'
    else:
        standard_str = ''
    fv_mode = 'foveated' if args.foveated else 'non-foveated'
    logging.info( '"FovVideoVDP v{}, {:.4g} [pix/deg], Lpeak={:.5g}, Lblack={:.4g} [cd/m^2], {}{}"'.format(fv.version, fv.pix_per_deg, fv.display_photometry.get_peak_luminance(), fv.display_photometry.get_black_level(), fv_mode, standard_str) )

    for kk in range( max(N_test, N_ref) ):
        test_file = args.test[min(kk,N_test-1)]
        ref_file = args.ref[min(kk,N_ref-1)]
        logging.info("Predicting the quality of '" + test_file + "' compared to '" + ref_file + "' ...")
        vs = pyfvvdp.fvvdp_video_source_file( test_file, ref_file )
        Q_jod, stats = fv.predict_video_source(vs)
        if args.quiet:                
            print( "{Q_jod:0.4f}".format(Q_jod=Q_jod) )
        else:
            print( "Q_JOD={Q_jod:0.4f}".format(Q_jod=Q_jod) )

        if do_heatmap:
            # diff_type = heatmap_types[args.heatmap]
            # heatmap = stats["heatmap"] * diff_type["scale"]
            # diff_map_viz = visualize_diff_map(heatmap, context_image=ref_vid_luminance, colormap_type=diff_type["colormap_type"])
            out_dir = "." if args.heatmap_dir is None else args.heatmap_dir
            os.makedirs(out_dir, exist_ok=True)
            base, ext = os.path.splitext(os.path.basename(test_file))            
            if stats["heatmap"].shape[2]>1: # if it is a video
                dest_name = os.path.join(out_dir, base + "_heatmap.mp4")
                logging.info("Writing heat map '" + dest_name + "' ...")
                np2vid(torch.squeeze(stats["heatmap"].permute((2,3,4,1,0)), dim=4).cpu().numpy(), dest_name, vs.get_frames_per_second(), args.verbose)
            else:
                dest_name = os.path.join(out_dir, base + "_heatmap.png")
                logging.info("Writing heat map '" + dest_name + "' ...")
                np2img(torch.squeeze(stats["heatmap"].permute((2,3,4,1,0)), dim=4).cpu().numpy(), dest_name)
                
            del stats

    #     del test_vid
    #     torch.cuda.empty_cache()

if __name__ == '__main__':
    main()



from abc import abstractmethod
from urllib.parse import ParseResultBytes
from numpy.lib.shape_base import expand_dims
import torch
from torch.functional import Tensor
import torch.nn.functional as Func
import numpy as np 
import os
import sys
import argparse
import time
import math
import torch.utils.benchmark as torchbench
import logging

from pyfvvdp.visualize_diff_map import visualize_diff_map

# For debugging only
# from gfxdisp.pfs.pfs_torch import pfs_torch

sys.path.append(os.path.dirname(os.path.abspath(__file__)))

from third_party.cpuinfo import cpuinfo
from hdrvdp_lpyr_dec import hdrvdp_lpyr_dec
from interp import interp1, interp3
from utils import *
from fvvdp_test import FovVideoVDP_Testbench

from pyfvvdp.fvvdp_display_model import fvvdp_display_photometry, fvvdp_display_geometry


"""
fvvdp_video_source_* objects are used to supply test/reference frames to FovVideoVDP. 
Those could be comming from memory or files. The subclasses of this abstract class implement
reading the frames and converting them to the approprtate format. 
"""
class fvvdp_video_source:
   
    # Return (height, width, frames) touple with the resolution and
    # the length of the video clip.
    @abstractmethod
    def get_video_size(self):
        pass

    # Return the frame rate of the video
    @abstractmethod
    def get_frames_per_second(self) -> int:
        pass
    
    # Get a pair of test and reference video frames as a single-precision luminance map
    # scaled in absolute inits of cd/m^2. 'frame' is the frame index,
    # starting from 0. 
    @abstractmethod
    def get_test_frame( self, frame, device ) -> Tensor:
        pass

    @abstractmethod
    def get_reference_frame( self, frame, device ) -> Tensor:
        pass


"""
Function for changing the order of dimensions, for example, from "WHC" (width, height, colour) to "BCHW" (batch, colour, height, width)
If a dimension is missing in in_dims, it will be added as a singleton dimension
"""
def reshuffle_dims( T: Tensor, in_dims: str, out_dims: str ) -> Tensor:
    in_dims = in_dims.upper()
    out_dims = out_dims.upper()    

    # Find intersection of two strings    
    inter_dims = ""
    for kk in range(len(out_dims)):
        if in_dims.find(out_dims[kk]) != -1:
            inter_dims += out_dims[kk]

    # First, permute into the right order
    perm = [0] * len(inter_dims)
    for kk in range(len(inter_dims)):
        ind = in_dims.find(inter_dims[kk])
        if ind == -1:
            raise RuntimeError( 'Dimension "{}" missing in the target dimensions: "{}"'.format(in_dims[kk],out_dims) )
        perm[kk] = ind                    
    T_p = T.permute(perm)

    # Add missing dimensions
    out_sh = [1] * len(out_dims)
    for kk in range(len(out_dims)):        
        ind = inter_dims.find(out_dims[kk])
        if ind != -1:
            out_sh[kk] = T_p.shape[ind]

    return T_p.reshape( out_sh )


"""
This video_source uses a photometric display model to convert input content (e.g. sRGB) to luminance maps. 
"""
class fvvdp_video_source_dm( fvvdp_video_source ):

    def __init__( self,  display_photometry='sdr_4k_30', color_space_name='sRGB' ):
        colorspaces_file = os.path.join(os.path.dirname(__file__), "fvvdp_data/color_spaces.json")
        colorspaces = json2dict(colorspaces_file)

        if not color_space_name in colorspaces:
            raise RuntimeError( "Unknown color space: \"" + color_space_name + "\"" )

        self.color_to_luminance = colorspaces[color_space_name]['RGB2Y']

        if isinstance( display_photometry, str ):
            display_models_file = os.path.join(os.path.dirname(__file__), "fvvdp_data/display_models.json")
            display_models = json2dict(display_models_file)

            self.dm_photometry = fvvdp_display_photometry.load(display_photometry)
        elif isinstance( display_photometry, fvvdp_display_photometry ):
            self.dm_photometry = display_photometry
        else:
            raise RuntimeError( "display_model must be a string or fvvdp_display_photometry subclass" )


"""
This video source supplies frames from either Pytorch tensors and Numpy arrays. It also applies a photometric display model.

A batch of videos should be stored as a tensor or numpy array. Ideally, the tensor should have the dimensions BCFHW (batch, colour, frame, height, width).If tensor is stored in another formay, you can pass the order of dimsions as "dim_order" parameter. If any dimension is missing, it will
be added as a singleton dimension. 

This class is for display-encoded (gamma-encoded) content that will be processed by a display model to produce linear  absolute luminance emitted from a display.
"""
class fvvdp_video_source_array( fvvdp_video_source_dm ):
               
    # test_video, reference video - tensor with test and reference video frames. See the class description above for the explanation of dimensions of those tensors.
    # fps - frames per second. Must be 0 for images
    # dim_order - a string with the order of the dimensions. 'BCFHW' is the default.
    # display_model - object that implements fvvdp_display_photometry
    #   class
    # color_space_name - name of the colour space (see
    #   fvvdp_data/color_spaces.json)
    def __init__( self, test_video, reference_video, fps, dim_order='BCFHW', display_photometry='sdr_4k_30', color_space_name='sRGB' ):

        super().__init__(display_photometry=display_photometry, color_space_name=color_space_name)        

        if test_video.shape != reference_video.shape:
            raise RuntimeError( 'Test and reference image/video tensors must be exactly the same shape' )
        
        if len(dim_order) != len(test_video.shape):
            raise RuntimeError( 'Input tensor much have exactly as many dimensions as there are characters in the "dims" parameter' )

        # Convert numpy arrays to tensors. Note that we do not upload to device or change dtype at this point (to save GPU memory)
        if isinstance( test_video, np.ndarray ):
            if test_video.dtype == np.uint16:
                # Torch does not natively support uint16. A workaround is to pack uint16 values into int16.
                # This will be efficiently transferred and unpacked on the GPU.
                # logging.info('Test has datatype uint16, packing into int16')
                test_video = test_video.astype(np.int16)
            test_video = torch.tensor(test_video)
        if isinstance( reference_video, np.ndarray ):
            if reference_video.dtype == np.uint16:
                # Torch does not natively support uint16. A workaround is to pack uint16 values into int16.
                # This will be efficiently transferred and unpacked on the GPU.
                # logging.info('Reference has datatype uint16, packing into int16')
                reference_video = reference_video.astype(np.int16)
            reference_video = torch.tensor(reference_video)

        # Change the order of dimension to match BFCHW - batch, frame, colour, height, width
        test_video = reshuffle_dims( test_video, in_dims=dim_order, out_dims="BCFHW" )
        reference_video = reshuffle_dims( reference_video, in_dims=dim_order, out_dims="BCFHW" )

        B, C, F, H, W = test_video.shape
        
        if fps==0 and F>1:
            raise RuntimeError( 'When passing video sequences, you must set ''frames_per_second'' parameter' )

        if C!=3 and C!=1:
            raise RuntimeError( 'The content must have either 1 or 3 colour channels.' )

        self.fps = fps
        self.is_video = (fps>0)
        self.is_color = (C==3)
        self.test_video = test_video
        self.reference_video = reference_video


    def get_frames_per_second(self):
        return self.fps
            
    # Return a [height width frames] vector with the resolution and
    # the number of frames in the video clip. [height width 1] is
    # returned for an image. 
    def get_video_size(self):

        sh = self.test_video.shape
        return (sh[3], sh[4], sh[2])
    
    # % Get a test video frame as a single-precision luminance map
    # % scaled in absolute inits of cd/m^2. 'frame' is the frame index,
    # % starting from 0. If use_gpu==true, the function should return a
    # % gpuArray.

    def get_test_frame( self, frame, device=torch.device('cpu') ):
        return self._get_frame(self.test_video, frame, device )

    def get_reference_frame( self, frame, device=torch.device('cpu') ):
        return self._get_frame(self.reference_video, frame, device )

    def _get_frame( self, from_array, frame, device ):        
        # Determine the maximum value of the data type storing the
        # image/video

        if from_array.dtype is torch.float32:
            frame = from_array[:,:,frame:(frame+1),:,:].to(device)
        elif from_array.dtype is torch.int16:
            # Use int16 to losslessly pack uint16 values
            # Unpack from int16 by bit masking as described in this thread:
            # https://stackoverflow.com/a/20766900
            # logging.info('Found int16 datatype, unpack into uint16')
            max_value = 2**16 - 1
            # Cast to int32 to store values >= 2**15
            frame_int32 = from_array[:,:,frame:(frame+1),:,:].to(device).to(torch.int32)
            frame_uint16 = frame_int32 & max_value
            # Finally convert to float in the range [0,1]
            frame = frame_uint16.to(torch.float32) / max_value
        elif from_array.dtype is torch.uint8:
            frame = from_array[:,:,frame:(frame+1),:,:].to(device).to(torch.float32)/255
        else:
            raise RuntimeError( "Only uint8, uint16 and float32 is currently supported" )

        L = self.dm_photometry.forward( frame )
        
        if self.is_color:
            # Convert to grayscale
            L = L[:,0:1,:,:,:]*self.color_to_luminance[0] + L[:,1:2,:,:,:]*self.color_to_luminance[1] + L[:,2:3,:,:,:]*self.color_to_luminance[2]

        return L

"""
FovVideoVDP metric. Refer to pytorch_examples for examples on how to use this class. 
"""
class fvvdp:
    def __init__(self, display_name="standard_4k", display_photometry=None, display_geometry=None, color_space="sRGB", foveated=False, heatmap=None, quiet=False, device=None):
        self.quiet = quiet
        self.foveated = foveated
        self.heatmap = heatmap
        self.color_space = color_space

        if display_photometry is None:
            self.display_photometry = fvvdp_display_photometry.load(display_name)
        else:
            self.display_photometry = display_photometry
        
        self.do_heatmap = (not self.heatmap is None) and (self.heatmap != "none")

        if display_geometry is None:
            self.display_geometry = fvvdp_display_geometry.load(display_name)
        else:
            self.display_geometry = display_geometry

        self.pix_per_deg = self.display_geometry.get_ppd()

        # Use GPU if available
        if device is None:
            if torch.cuda.is_available() and torch.cuda.device_count()>0:
                self.device = torch.device('cuda:0')
            else:
                self.device = torch.device('cpu')
        else:
            self.device = device

        self.load_config()

        # if self.mask_s > 0.0:
        #     self.mask_p = self.mask_q + self.mask_s

        self.csf_cache              = {}
        self.csf_cache_dirs         = [
                                        "csf_cache",
                                        os.path.join(os.path.dirname(__file__), "csf_cache"),
                                      ]

        self.omega = torch.tensor([0,5], device=self.device, requires_grad=False)
        for oo in self.omega:
            self.preload_cache(oo, self.csf_sigma)

        self.lpyr = None
        self.imgaussfilt = ImGaussFilt(0.5 * self.pix_per_deg, self.device)


    def load_config( self ):

        parameters_file = os.path.join(os.path.dirname(__file__), "fvvdp_data/fvvdp_parameters.json")
        parameters = json2dict(parameters_file)

        #all common parameters between Matlab and Pytorch, loaded from the .json file
        self.mask_p = parameters['mask_p']
        self.mask_c = parameters['mask_c'] # content masking adjustment
        self.pu_dilate = parameters['pu_dilate']
        self.w_transient = parameters['w_transient'] # The weight of the transient temporal channel
        self.beta = parameters['beta'] # The exponent of the spatial summation (p-norm)
        self.beta_t = parameters['beta_t'] # The exponent of the summation over time (p-norm)
        self.beta_tch = parameters['beta_tch'] # The exponent of the summation over temporal channels (p-norm)
        self.beta_sch = parameters['beta_sch'] # The exponent of the summation over spatial channels (p-norm)
        self.sustained_sigma = parameters['sustained_sigma']
        self.sustained_beta = parameters['sustained_beta']
        self.csf_sigma = parameters['csf_sigma']
        self.sensitivity_correction = parameters['sensitivity_correction'] # Correct CSF values in dB. Negative values make the metric less sensitive.
        self.masking_model = parameters['masking_model']
        self.local_adapt = parameters['local_adapt'] # Local adaptation: 'simple' or or 'gpyr'
        self.contrast = parameters['contrast']  # Either 'weber' or 'log'
        self.jod_a = parameters['jod_a']
        self.log_jod_exp = parameters['log_jod_exp']
        self.mask_q_sust = parameters['mask_q_sust']
        self.mask_q_trans = parameters['mask_q_trans']
        self.k_cm = parameters['k_cm']  # new parameter controlling cortical magnification
        self.filter_len = parameters['filter_len']
        self.version = parameters['version']

        # other parameters
        self.debug = False

    '''
    Predict image/video quality using FovVideoVDP.

    test_cont and reference_cont can be either numpy arrays or PyTorch tensors with images or video frames. 
        Depending on the display model (display_photometry), the pixel values should be either display encoded, or absolute linear.
        The two supported datatypes are float16 and uint8.
    dim_order - a string with the order of dimensions of test_cont and reference_cont. The individual characters denote
        B - batch
        C - colour channel
        F - frame
        H - height
        W - width
        Examples: "HW" - gray-scale image (column-major pixel order); "HWC" - colour image; "FCHW" - colour video
        The default order is "BCFHW". The processing can be a bit faster if data is provided in that order. 
    frame_padding - the metric requires at least 250ms of video for temporal processing. Because no previous frames exist in the
        first 250ms of video, the metric must pad those first frames. This options specifies the type of padding to use:
          'replicate' - replicate the first frame
          'circular'  - tile the video in the front, so that the last frame is used for frame 0.
          'pingpong'  - the video frames are mirrored so that frames -1, -2, ... correspond to frames 0, 1, ...
    '''
    def predict(self, test_cont, reference_cont, dim_order="BCFHW", frames_per_second=0, fixation_point=None, frame_padding="replicate"):

        test_vs = fvvdp_video_source_array( test_cont, reference_cont, frames_per_second, dim_order=dim_order, display_photometry=self.display_photometry, color_space_name=self.color_space )

        return self.predict_video_source(test_vs, fixation_point=fixation_point, frame_padding=frame_padding)

    '''
    The same as `predict` but takes as input fvvdp_video_source_* object instead of Numpy/Pytorch arrays.
    '''
    def predict_video_source(self, vid_source, fixation_point=None, frame_padding="replicate"):

        # T_vid and R_vid are the tensors of the size (1,1,N,H,W)
        # where:
        # N - the number of frames
        # H - height in pixels
        # W - width in pixels
        # Both images must contain linear absolute luminance values in cd/m^2
        # 
        # We assume the pytorch default NCDHW layout

        vid_sz = vid_source.get_video_size() # H, W, F
        height, width, N_frames = vid_sz

        if fixation_point is None:
            fixation_point = torch.tensor([width//2, height//2])
        elif isinstance(fixation_point, np.ndarray):
            fixation_point = torch.tensor(fixation_point)

        if self.lpyr is None or self.lpyr.W!=width or self.lpyr.H!=height:
            self.lpyr = hdrvdp_lpyr_dec(width, height, self.pix_per_deg, self.device)

        #assert self.W == R_vid.shape[-1] and self.H == R_vid.shape[-2]
        #assert len(R_vid.shape)==5

        is_image = (N_frames==1)  # Can run faster on images

        if is_image:
            temp_ch = 1  # How many temporal channels
        else:
            temp_ch = 2
            self.filter_len = int(np.ceil( 250.0 / (1000.0/vid_source.get_frames_per_second()) ))

            self.F, self.omega = self.get_temporal_filters(vid_source.get_frames_per_second())
            # if self.debug: self.tb.verify_against_matlab(self.F,     'F_pod', self.device)
            # if self.debug: self.tb.verify_against_matlab(self.omega, 'omega', self.device)


        if self.do_heatmap:
            dmap_channels = 1 if self.heatmap == "raw" else 3
            heatmap = torch.zeros([1,dmap_channels,N_frames,height,width], dtype=torch.float, device=self.device) 
        else:
            heatmap = None

        N_nCSF = []
        sw_buf = [None, None]
        Q_per_ch = None
        w_temp_ch = [1.0, self.w_transient]

        fl = self.filter_len

        for ff in range(N_frames):

            if is_image:                
                R = torch.zeros((1, 2, 1, height, width), device=self.device)
                R[:,0, :, :, :] = vid_source.get_test_frame(0, device=self.device)
                R[:,1, :, :, :] = vid_source.get_reference_frame(0, device=self.device)

            else: # This is video
                if self.debug: print("Frame %d:\n----" % ff)

                if ff == 0: # First frame
                    if frame_padding == "replicate":
                        # TODO: proper handling of batches
                        sw_buf[0] = vid_source.get_test_frame(0, device=self.device).expand([1, 1, fl, height, width])
                        sw_buf[1] = vid_source.get_reference_frame(0, device=self.device).expand([1, 1, fl, height, width])
                    elif frame_padding == "circular":
                        sw_buf[0] = torch.zeros([1, 1, fl, height, width], device=self.device)
                        sw_buf[1] = torch.zeros([1, 1, fl, height, width], device=self.device)
                        for kk in range(fl):
                            fidx = (N_frames - 1 - fl + kk) % N_frames
                            sw_buf[0][:,:,kk,...] = vid_source.get_test_frame(fidx, device=self.device)
                            sw_buf[1][:,:,kk,...] = vid_source.get_reference_frame(fidx, device=self.device)
                    elif frame_padding == "pingpong":
                        sw_buf[0] = torch.zeros([1, 1, fl, height, width], device=self.device)
                        sw_buf[1] = torch.zeros([1, 1, fl, height, width], device=self.device)

                        pingpong = list(range(0,N_frames)) + list(range(N_frames-2,0,-1))
                        indices = []
                        while(len(indices) < (fl-1)):
                            indices = indices + pingpong
                        indices = indices[-(fl-1):] + [0]

                        for kk in range(fl):
                            fidx = indices[kk]
                            sw_buf[0][:,:,kk,...] = vid_source.get_test_frame(fidx,device=self.device)
                            sw_buf[1][:,:,kk,...] = vid_source.get_reference_frame(fidx,device=self.device)
                    else:
                        raise RuntimeError( 'Unknown padding method "{}"'.format(frame_padding) )
                else:
                    cur_tframe = vid_source.get_test_frame(ff, device=self.device)
                    cur_rframe = vid_source.get_reference_frame(ff, device=self.device)

                    sw_buf[0] = torch.cat((sw_buf[0][:, :, 1:, :, :], cur_tframe), 2)
                    sw_buf[1] = torch.cat((sw_buf[1][:, :, 1:, :, :], cur_rframe), 2)

                # Order: test-sustained, ref-sustained, test-transient, ref-transient
                R = torch.zeros((1, 4, 1, height, width), device=self.device)

                for cc in range(temp_ch):
                    # 1D filter over time (over frames)
                    corr_filter = self.F[cc].flip(0).view([1,1,self.F[cc].shape[0],1,1]) 
                    R[:,cc*2+0, :, :, :] = (sw_buf[0] * corr_filter).sum(dim=-3,keepdim=True)
                    R[:,cc*2+1, :, :, :] = (sw_buf[1] * corr_filter).sum(dim=-3,keepdim=True)

            if self.debug: self.tb.verify_against_matlab(R.permute(0,2,3,4,1), 'Rdata', self.device, file='R_%d' % (ff+1), tolerance = 0.01)

            # Perform Laplacian decomposition
            # B = [None] * R.shape[1]
            # for rr in range(R.shape[1]):
            #     B[rr] = hdrvdp_lpyr_dec( R[0,rr:(rr+1),0:1,:,:], self.pix_per_deg, self.device)

            B_bands, B_gbands = self.lpyr.decompose(R[0,...])

            if self.debug: assert len(B_bands) == self.lpyr.get_band_count()

            # CSF
            N_nCSF = [[None, None] for i in range(self.lpyr.get_band_count()-1)]

            if self.do_heatmap:
                Dmap_pyr_bands, Dmap_pyr_gbands = self.lpyr.decompose( torch.zeros([1,1,height,width], dtype=torch.float, device=self.device))

            # L_bkg_bb = [None for i in range(self.lpyr.get_band_count()-1)]

            rho_band = self.lpyr.get_freqs()

            # Adaptation
            L_adapt = None

            if self.local_adapt == "simple":
                L_adapt = R[0,1,0,...] # reference, sustained
                if self.contrast == "log":
                    L_adapt = torch.pow(10.0, L_adapt)
                L_adapt = self.imgaussfilt.run(L_adapt)
            elif self.local_adapt == "global":
                print("ERROR: global adapt not supported")
                return

            for cc in range(temp_ch):
                for bb in range(self.lpyr.get_band_count()-1):

                    T_f = self.lpyr.get_band(B_bands, bb)[cc*2+0,0,...]
                    R_f = self.lpyr.get_band(B_bands, bb)[cc*2+1,0,...]

                    L_bkg, R_f, T_f = self.compute_local_contrast(R_f, T_f, 
                        self.lpyr.get_gband(B_gbands, bb+1)[1:2,...], L_adapt)

                    # temp_errs[ff] += torch.mean(torch.abs(R_f - T_f))
                    # continue

                    # if self.debug: self.tb.verify_against_matlab(L_bkg, 'L_bkg_data', self.device, file='L_bkg_%d_%d_%d' % (ff+1,bb+1,cc+1), tolerance = 0.01)

                    # #  TODO: check cc = 1
                    # if self.debug:
                    #     print("%d, %d" % (cc, bb))
                    #     h, w = T_f.shape[-2], T_f.shape[-1]
                    #     np2img(stack_horizontal([
                    #         (T_f * 100.0).squeeze().unsqueeze(-1).expand(h,w,3).cpu().numpy(), 
                    #         (R_f * 100.0).squeeze().unsqueeze(-1).expand(h,w,3).cpu().numpy()])).show()
                    if self.debug: self.tb.verify_against_matlab(T_f, 'T_f_data', self.device, file='T_f_%d_%d_%d' % (ff+1,bb+1,cc+1), tolerance = 0.001) 
                    if self.debug: self.tb.verify_against_matlab(R_f, 'R_f_data', self.device, file='R_f_%d_%d_%d' % (ff+1,bb+1,cc+1), tolerance = 0.001)

                    if N_nCSF[bb][cc] is None:

                        if self.foveated:   # Fixation, parafoveal sensitivity
                            if fixation_point.dim() == 2:
                                current_fixation_point = fixation_point[ff,:].squeeze()
                            else:
                                current_fixation_point = fixation_point

                            xv = torch.linspace( 0.5, vid_sz[1]-0.5, T_f.shape[-1], device=self.device )
                            yv = torch.linspace( 0.5, vid_sz[0]-0.5, T_f.shape[-2], device=self.device )
                            [xx, yy] = torch.meshgrid( xv, yv, indexing='xy' )

                            ecc = self.display_geometry.pix2eccentricity( torch.tensor((vid_sz[1], vid_sz[0])), xx, yy, current_fixation_point+0.5 )

                            # The same shape as bands
                            ecc = ecc.reshape( [1, 1, ecc.shape[-2], ecc.shape[-1]] )

                            res_mag = self.display_geometry.get_resolution_magnification(ecc)
                        else:   # No fixation, foveal sensitivity everywhere
                            res_mag = torch.ones(R_f.shape[-2:], device=self.device)
                            ecc = torch.zeros(R_f.shape[-2:], device=self.device)

                        rho = rho_band[bb] * res_mag

                        # if self.debug: self.tb.verify_against_matlab(ecc, 'ecc_data', self.device, file='ecc_%d_%d_%d' % (ff+1,bb+1,cc+1), tolerance = 0.01)  
                        # if self.debug: self.tb.verify_against_matlab(rho, 'rho_data', self.device, file='rho_%d_%d_%d' % (ff+1,bb+1,cc+1), tolerance = 0.05)  

                        S = self.cached_sensitivity(rho, self.omega[cc], L_bkg, ecc, self.csf_sigma) * np.power(10.0, self.sensitivity_correction/20.0)
                        # if self.debug: self.tb.verify_against_matlab(S, 'S_data', self.device, file='S_%d_%d_%d' % (ff+1,bb+1,cc+1), tolerance = 0.01, verbose=False)

                        if self.contrast == "log": N_nCSF[bb][cc] = self.weber2log(torch.min(1./S, self.torch_scalar(0.9999999)))
                        else:                      N_nCSF[bb][cc] = torch.reciprocal(S)
                        # if self.debug: self.tb.verify_against_matlab(N_nCSF[bb][cc], 'N_nCSF_data', self.device, file='N_nCSF_%d_%d_%d' % (ff+1,bb+1,cc+1), tolerance = 0.01, relative = True)

                    D = self.apply_masking_model(T_f, R_f, N_nCSF[bb][cc], cc)
                    if self.debug: self.tb.verify_against_matlab(D, 'D_data', self.device, file='D_%d_%d_%d' % (ff+1,bb+1,cc+1), tolerance = 0.1, relative = True, verbose=False)

                    if self.do_heatmap:
                        if cc == 0: self.lpyr.set_band(Dmap_pyr_bands, bb, D)
                        else:       self.lpyr.set_band(Dmap_pyr_bands, bb, self.lpyr.get_band(Dmap_pyr_bands, bb) + w_temp_ch[cc] * D)

                    if Q_per_ch is None:
                        Q_per_ch = torch.zeros((len(B_bands)-1, 2, N_frames), device=self.device)

                    Q_per_ch[bb,cc,ff] = w_temp_ch[cc] * self.lp_norm(D.flatten(), self.beta, 0, True)

                    if self.debug: self.tb.verify_against_matlab(Q_per_ch[bb,cc,ff], 'Q_per_ch_data', self.device, file='Q_per_ch_%d_%d_%d' % (bb+1,cc+1,ff+1), tolerance = 0.1, relative=True, verbose=False)
            # break
            if self.do_heatmap:
                beta_jod = np.power(10.0, self.log_jod_exp)
                dmap = torch.pow(self.lpyr.reconstruct(Dmap_pyr_bands), beta_jod) * abs(self.jod_a)         
                if self.heatmap == "raw":
                    heatmap[:,:,ff,...] = dmap 
                else:
                    ref_frame = R[:,0, :, :, :]
                    heatmap[:,:,ff,...] = visualize_diff_map(dmap, context_image=ref_frame, colormap_type=self.heatmap)


        Q_sc = self.lp_norm(Q_per_ch, self.beta_sch, 0, False)
        Q_tc = self.lp_norm(Q_sc,     self.beta_tch, 1, False)
        Q    = self.lp_norm(Q_tc,     self.beta_t,   2, True)

        Q = Q.squeeze()

        beta_jod = np.power(10.0, self.log_jod_exp)
        Q_jod = np.sign(self.jod_a) * torch.pow(torch.tensor(np.power(np.abs(self.jod_a),(1.0/beta_jod)), device=self.device)* Q, beta_jod) + 10.0 # This one can help with very large numbers

        stats = {}

        if self.do_heatmap:            
            stats['heatmap'] = heatmap

        if self.debug: self.tb.verify_against_matlab(Q_per_ch, 'Q_per_ch_data', self.device, file='Q_per_ch', tolerance = 0.1, relative=True, verbose=True)
        if self.debug: self.tb.verify_against_matlab(Q_sc,     'Q_sc_data',     self.device, file='Q_sc',     tolerance = 0.1, relative=True, verbose=True)
        if self.debug: self.tb.verify_against_matlab(Q_tc,     'Q_tc_data',     self.device, file='Q_tc',     tolerance = 0.1, relative=True, verbose=True)
        if self.debug: self.tb.verify_against_matlab(Q,        'Q_data',        self.device, file='Q',        tolerance = 0.1, relative=True, verbose=True)
        if self.debug: self.tb.verify_against_matlab(Q_jod,    'Q_jod_data',    self.device, file='Q_jod',    tolerance = 0.1, relative=True, verbose=True)
        if self.debug: self.tb.print_summary()    

        return (Q_jod.squeeze(), stats)

    def compute_local_contrast(self, R, T, next_gauss_band, L_adapt):
        if self.local_adapt=="simple":
            L_bkg = Func.interpolate(L_adapt.unsqueeze(0).unsqueeze(0), R.shape, mode='bicubic', align_corners=True)
            # L_bkg = torch.ones_like(R) * torch.mean(R)
            # np2img(l2rgb(L_adapt.unsqueeze(-1).cpu().data.numpy())/200.0).show()
            # np2img(l2rgb(L_bkg[0,0].unsqueeze(-1).cpu().data.numpy())/200.0).show()
        elif self.local_adapt=="gpyr":
            if self.contrast == "log":
                next_gauss_band = torch.pow(10.0, next_gauss_band)
            L_bkg = self.lpyr.gausspyr_expand(next_gauss_band, [R.shape[-2], R.shape[-1]])
        else:
            print("Error: local adaptation %s not supported" % self.local_adapt)
            return

        if self.contrast != "log":
            L_bkg_clamped = torch.clamp(L_bkg, min=0.1)
            T = torch.clamp(torch.div(T, L_bkg_clamped), max=1000.0)
            R = torch.clamp(torch.div(R, L_bkg_clamped), max=1000.0)

        return L_bkg, R, T

    def get_cache_key(self, omega, sigma, k_cm):
        return ("o%g_s%g_cm%f" % (omega, sigma, k_cm)).replace('-', 'n').replace('.', '_')

    def preload_cache(self, omega, sigma):
        key = self.get_cache_key(omega, sigma, self.k_cm)
        for csf_cache_dir in self.csf_cache_dirs:
            #fname = os.path.join(csf_cache_dir, key + '_cpu.mat')
            fname = os.path.join(csf_cache_dir, key + '_gpu0.mat')
            if os.path.isfile(fname):
                #lut = load_mat_dict(fname, "lut_cpu", self.device)
                lut = load_mat_dict(fname, "lut", self.device)
                for k in lut:
                    lut[k] = torch.tensor(lut[k], device=self.device, requires_grad=False)
                self.csf_cache[key] = {"lut" : lut}
                break
        if key not in self.csf_cache:
            raise RuntimeError("Error: cache file for %s not found" % key)

    def cached_sensitivity(self, rho, omega, L_bkg, ecc, sigma):
        key = self.get_cache_key(omega, sigma, self.k_cm)

        if key in self.csf_cache:
            lut = self.csf_cache[key]["lut"]
        else:
            print("Error: Key %s not found in cache" % key)

        # ASSUMPTION: rho_q and ecc_q are not scalars
        rho_q = torch.log2(torch.clamp(rho,   lut["rho"][0], lut["rho"][-1]))
        Y_q   = torch.log2(torch.clamp(L_bkg, lut["Y"][0],   lut["Y"][-1]))
        ecc_q = torch.sqrt(torch.clamp(ecc,   lut["ecc"][0], lut["ecc"][-1]))

        interpolated = interp3( lut["rho_log"], lut["Y_log"], lut["ecc_sqrt"], lut["S_log"], rho_q, Y_q, ecc_q)

        S = torch.pow(2.0, interpolated)

        return S

    def weber2log(self, W):
        # Convert Weber contrast 
        #
        # W = (B-A)/A
        #
        # to log contrast
        #
        # G = log10( B/A );
        #
        return torch.log10(1.0 + W)

    def phase_uncertainty(self, M):
        if self.pu_dilate != 0:
            M_pu = imgaussfilt( M, self.pu_dilate ) * torch.pow(10.0, self.mask_c)
        else:
            M_pu = M * torch.pow(self.torch_scalar(10.0), self.torch_scalar(self.mask_c))
        return M_pu

    def mask_func_perc_norm(self, G, G_mask):
        # Masking on perceptually normalized quantities (as in Daly's VDP)
        
        p = self.mask_p
        #q = self.mask_q
        #k = self.k_mask_self
        
        R = torch.div(torch.pow(G,p), torch.sqrt( 1. + torch.pow(k * G_mask, 2*q)))

        return R

    def mask_func_perc_norm2(self, G, G_mask, p, q ):
        # Masking on perceptually normalized quantities (as in Daly's VDP)        
        R = torch.div(torch.pow(G,p), 1. + torch.pow(G_mask, q))
        return R

    def apply_masking_model(self, T, R, N, cc):
        # cc - temporal channel: 0 - sustained, 1 - transient
        if self.masking_model == "joint_mutual_masking_perc_norm":
            T = torch.div(T, N)
            R = torch.div(R, N)
            M = self.phase_uncertainty( torch.min( torch.abs(T), torch.abs(R) ) )
            d = torch.abs(T-R)
            M = M + d
            D = self.mask_func_perc_norm( d, M )
        elif self.masking_model == 'min_mutual_masking_perc_norm2':
            p = self.mask_p
            if cc==0:
                q = self.mask_q_sust
            else:
                q = self.mask_q_trans     
            T = torch.div(T, N)
            R = torch.div(R, N)
            M = self.phase_uncertainty( torch.min( torch.abs(T), torch.abs(R) ) )
            D = self.mask_func_perc_norm2( torch.abs(T-R), M, p, q )
        else:
            print("Error: Masking model" + self.masking_model + "not implemented")

        D = torch.clamp( D, max=1e4)
        return D

    def lp_norm(self, x, p, dim=0, normalize=True):
        if dim is None:
            dim = 0

        if normalize:
            N = x.shape[dim]
        else:
            N = 1.0

        return torch.norm(x, p, dim=dim, keepdim=True) / (float(N) ** (1./p))

    def get_temporal_filters(self, frames_per_s):
        t = torch.linspace(0.0, self.filter_len / frames_per_s, self.filter_len, device=self.device)
        F = torch.zeros((2, t.shape[0]), device=self.device)

        sigma = torch.tensor([self.sustained_sigma], device=self.device)
        beta = torch.tensor([self.sustained_beta], device=self.device)

        F[0] = torch.exp(-torch.pow(torch.log(t + 1e-4) - torch.log(beta), 2.0) / (2.0 * (sigma ** 2.0)))
        F[0] = F[0] / torch.sum(F[0])

        Fdiff = F[0, 1:] - F[0, :-1]

        k2 = 0.062170507756932
        # This one seems to be slightly more accurate at low sampling rates
        F[1] = k2*torch.cat([Fdiff/(t[1]-t[0]), torch.tensor([0.0], device=self.device)], 0) # transient

        omega = torch.tensor([0,5], device=self.device, requires_grad=False)

        F[0].requires_grad = False
        F[1].requires_grad = False

        return F, omega

    def torch_scalar(self, val, dtype=torch.float32):
        return torch.tensor(val, dtype=dtype, device=self.device)
        

# def parse_args():
#     parser = argparse.ArgumentParser(description="Video-VDP metric test app")
#     parser.add_argument("--gpu",      type=int,  default=-1, help="use GPU")
#     group = parser.add_mutually_exclusive_group(required = True)
#     group.add_argument("--benchmark", type=str, default=None, required=False, help="benchmark performance for chosen resolution: e.g. 1920x1080x60")
#     group.add_argument("--test", action='store_true', default=False, required=False, help="test against FovDots MATLAB dataset (must be present on disk)")

#     return parser.parse_args()

# if __name__ == '__main__':

#     torch.set_printoptions(precision=4, sci_mode=False, linewidth=120)

#     args = parse_args()

#     if args.gpu >= 0 and torch.cuda.is_available():
#         device = torch.device('cuda:' + str(args.gpu))
#     else:
#         device = torch.device('cpu')

#     print("Running on " + str(device))

#     if args.benchmark is None:

#         n_contents = 18
#         n_conditions = 9

#         test_content = 14 # 15 in MATLAB
#         test_condition = 7 # 8 in MATLAB

#         H = 1600
#         W = 1400

#         vdp = FovVideoVDP( H=H, W=W, display_model=DisplayModel.load('HTC Vive Pro', 'sRGB'), frames_per_s=90.0, do_diff_map=False, device=device, debug=True)

#         for i_content in range(n_contents):
#             for i_condition in range(n_conditions):
#                 if i_content == test_content and i_condition == test_condition:

#                     print("%d-%d" % (i_content+1, i_condition+1))
#                     print("    loading")
#                     ref = fovdots_load_ref(i_content, device, data_res="full").to(device)
#                     test = fovdots_load_condition(i_content, i_condition, device, data_res="full").to(device)

#                     start_time = time.time()
#                     vdp.compute_metric(test, ref)
#                     elapsed_time = time.time() - start_time
#                     print("elapsed: %0.4f sec" % (elapsed_time))
#     else:
#         W, H, D = [int(x) for x in args.benchmark.split("x")]

#         print("CPU: ")
#         print(cpuinfo.cpu.info[0])

#         print("GPU: ")
#         print(torch.cuda.get_device_name())

#         with torch.no_grad():
#             vdp = FovVideoVDP(H=H, W=W, display_model=DisplayModel.load('HTC Vive Pro', 'sRGB'), frames_per_s=90.0, do_diff_map=False, device=device, debug=False)

#             ref  = torch.rand(1,1,D,H,W, device=device)
#             test = torch.rand(1,1,D,H,W, device=device)

#             bench = torchbench.Timer(stmt='vdp.compute_metric(ref, test)', globals=globals())

#             print("Torch Benchmark:")
#             print(bench.timeit(100))


# class TestLoss(torch.nn.Module):
#     def __init__(self, display_model, device, epsilon):
        
#         super(TestLoss, self).__init__()

#         self.display_model = display_model
#         self.device        = device
#         self.pix_per_deg   = self.display_model.get_ppd(0.0)
#         self.epsilon       = epsilon

#         mode = "blur2_adapt"

#         self.mode          = mode

#         if mode == "blur1_adapt":
#             self.imgaussfilt   = ImGaussFilt(1.0 * self.pix_per_deg, self.device)
#         elif mode == "blur2_adapt":
#             self.imgaussfilt   = ImGaussFilt(0.5 * self.pix_per_deg, self.device)



#         if mode is None:
#             logging.error("Error: Need mode")
#             sys.exit(1)

#         logging.info("Test loss mode: %s" % mode)

#     def forward(self, output, target):
#         N,C,D,H,W = target.shape

#         lossvals = torch.zeros((D,))

#         for ff in range(D):
#             T = target[:,:,ff,...]
#             R = output[:,:,ff,...]

#             if   self.mode == "no_adapt":    L_bkg = torch.ones_like(R)
#             elif self.mode == "mean_adapt":  L_bkg = torch.clamp(torch.ones_like(R) * torch.mean(R), min=self.epsilon)
#             elif self.mode == "blur1_adapt": L_bkg = torch.clamp(self.imgaussfilt.run(R),            min=self.epsilon)
#             elif self.mode == "blur2_adapt": L_bkg = torch.clamp(self.imgaussfilt.run(R),            min=self.epsilon)
#             elif self.mode == "mape_adapt":  L_bkg = torch.clamp(R,                                  min=self.epsilon)
#             elif self.mode == "smape_adapt": L_bkg = torch.clamp((R+T) * 0.5,                        min=self.epsilon)

#             lossvals[ff] = torch.mean(torch.abs(R/L_bkg - T/L_bkg))

#         return torch.mean(lossvals), None, None



from abc import ABC, abstractmethod
import torch
import torch.nn.functional as Func
import numpy as np 
import math
import logging
import os

from utils import json2dict

# Convert pixel values to linear RGB using sRGB non-linearity
# 
# L = srgb2lin( p )
#
# p - pixel values (between 0 and 1)
# L - relative linear RGB (or luminance), normalized to the range 0-1
def srgb2lin( p ):
    L = torch.where(p > 0.04045, ((p + 0.055) / 1.055)**2.4, p/12.92)
    return L

class fvvdp_display_photometry:

    # Transforms gamma-encoded pixel values V, which must be in the range
    # 0-into absolute linear colorimetric values emitted from
    # the display.
    @abstractmethod
    def forward( self, V ):
        pass

    # Print the display specification    
    @abstractmethod
    def print( self ):
        pass

    @classmethod
    def load( cls, display_name ):
        models_file      = os.path.join(os.path.dirname(__file__), "fvvdp_data/display_models.json")

        models = json2dict(models_file)

        if not display_name in models:
            raise RuntimeError( "Unknown display model: \"" + display_name + "\"" )

        model = models[display_name]

        Y_peak = model["max_luminance"]

        if "min_luminance" in model:
            contrast = Y_peak/model["min_luminance"]
        elif "contrast" in model:
            contrast = model["contrast"]
        else:
            contrast = 500

        # Ambient light
        if "E_ambient" in model:
            E_ambient = model["E_ambient"]
        else:
            E_ambient = 0
        
        # Reflectivity of the display panel
        if "k_refl" in model: 
            k_refl = model["k_refl"]
        else:
            k_refl = 0.005

        # Reflectivity of the display panel
        if "gamma" in model: 
            gamma = model["gamma"]
        else:
            gamma = 2.2

        obj = fvvdp_display_photo_gog( Y_peak, contrast, gamma, E_ambient, k_refl)
        obj.full_name = model["name"]
        obj.short_name = display_name

        return obj




class fvvdp_display_photo_gog(fvvdp_display_photometry): 
    # Gain-gamma-offset display model to simulate SDR displays
    # Object variables:
    #  Y_peak - display peak luminance in cd/m^2
    #  contrast - display contrast 
    #  gamma
    #  E_ambient
    #  k_refl
    
    # Gain-gamma-offset display model to simulate SDR displays
    #
    # dm = fvvdp_display_photo_gog( Y_peak )
    # dm = fvvdp_display_photo_gog( Y_peak, contrast )
    # dm = fvvdp_display_photo_gog( Y_peak, contrast, gamma )
    # dm = fvvdp_display_photo_gog( Y_peak, contrast, gamma, E_ambient )
    # dm = fvvdp_display_photo_gog( Y_peak, contrast, gamma, E_ambient, k_refl )
    #
    # Parameters (default value shown in []):
    # Y_peak - display peak luminance in cd/m^2 (nit), e.g. 200 for a typical
    #          office monitor
    # contrast - [1000] the contrast of the display. The value 1000 means
    #          1000:1
    # gamma - [-1] gamma of the display, typically 2.2. If -1 is
    #         passed, sRGB non-linearity is used.         
    # E_ambient - [0] ambient light illuminance in lux, e.g. 600 for bright
    #         office
    # k_refl - [0.005] reflectivity of the display screen
    #
    # For more details on the GOG display model, see:
    # https://www.cl.cam.ac.uk/~rkm38/pdfs/mantiuk2016perceptual_display.pdf
    #
    # Copyright (c) 2010-2021, Rafal Mantiuk
    def __init__( self, Y_peak, contrast = 1000, gamma = 2.2, E_ambient = 0, k_refl = 0.005 ):
            
        self.Y_peak = Y_peak            
        self.contrast = contrast
        self.gamma = gamma
        self.E_ambient = E_ambient
        self.k_refl = k_refl
    
        
    # Transforms gamma-encoded pixel values V, which must be in the range
    # 0-into absolute linear colorimetric values emitted from
    # the display.
    def forward( self, V ):
        
        if torch.any(V>1).bool() or torch.any(V<0).bool():
            logging.warning("Pixel outside the valid range 0-1")
            V = V.clamp( 0., 1. )
            
        Y_black = self.get_black_level()
        
        if self.gamma==-1: # sRGB
            L = (self.Y_peak-Y_black)*srgb2lin(V) + Y_black
        else:
            L = (self.Y_peak-Y_black)*torch.pow(V, self.gamma) + Y_black
        
        return L
        

    def get_peak_luminance( self ):
        return self.Y_peak


    # Get the effective black level, accounting for screen reflections
    def get_black_level( self ):
        Y_refl = self.E_ambient/math.pi*self.k_refl  # Reflected ambient light            
        Y_black = Y_refl + self.Y_peak/self.contrast

        return Y_black

    # Print the display specification    
    def print( self ):
        Y_black = self.get_black_level()
        
        logging.info( 'Photometric display model:' )
        logging.info( '  Peak luminance: {} cd/m^2'.format(self.Y_peak) )
        logging.info( '  Contrast - theoretical: {}:1'.format( math.round(self.contrast) ) )
        logging.info( '  Contrast - effective: {}:1'.format, math.round(self.Y_peak/Y_black) )
        logging.info( '  Ambient light: {} lux'.format( self.E_ambient ) )
        logging.info( '  Display reflectivity: {}%'.format( self.k_refl*100 ) )
    

class fvvdp_display_photo_absolute(fvvdp_display_photometry):
    # Use this photometric model when passing absolute colorimetric of
    # photometric values, scaled in cd/m^2
    # Object variables:
    #  L_max - display peak luminance in cd/m^2
    #  L_min - display black level
    def __init__(self, L_max=10000, L_min=0.005):

        self.L_max = L_max
        self.L_min = L_min


    def forward( self, V ):

        # Clamp the values that are outside the (L_min, L_max) range.
        L = V.clamp(self.L_min, self.L_max)

        if V.max() < 1:
            logging.warning('Pixel values are very low. Perhaps images are' \
                            ' not scaled in the absolute units of cd/m^2.')

        return L


    def  get_peak_luminance( self ):
        return self.L_max


    def get_black_level( self ):
        return self.L_min

    # Print the display specification
    def print( self ):
        Y_black = self.get_black_level()

        logging.info('Photometric display model:')
        logging.info('  Absolute photometric/colorimetric values')


# Use this class to compute the effective resolution of a display in pixels
# per degree (ppd). The class accounts for the change in the projection
# when looking at large FOV displays (e.g. VR headsets) at certain
# eccentricity.
#
# The class is also useful for computing the size of a display in meters
# and visual degrees. Check 'display_size_m' and 'display_size_deg' class
# properties for that.
#
# R = fvvdp_display_geometry(resolution, distance_m=None, distance_display_heights=None, 
#                            fov_horizontal=None, fov_vertical=None, fov_diagonal=None, 
#                            diagonal_size_inches=None)
#
# resolution is the 2-element touple with the pixel resolution of the
# display: (horizontal_resolutution, vertical_resolutution)
# distance_m - viewing distance in meters
# distance_display_heights - viewing distance in the heights of a display
# fov_horizontal - horizontal field of view of the display in degrees
# fov_vertical - vertical field of view of the display in degrees
# fov_diagonal - diagonal field of view of the display in degrees
# diagonal_size_inches - display diagonal resolution in inches
#
# Examples:
# # HTC Pro
# # Note that the viewing distance must be specified even though the resolution
# # and 'fov_diagonal' are enough to find pix_per_deg.
# R = fvvdp_display_geometry( (1440, 1600), distance_m=3, fov_diagonal=110 )
# R.get_ppd( torch.tensor( [0, 10, 20, 30, 40])) # pix per deg at the given eccentricities
#
# # 30" 4K monitor seen from 0.6 meters
# R = fvvdp_display_geometry( (3840, 2160), diagonal_size_inches=30, distance_m=0.6 )
# R.get_ppd()
#
# # 47" SIM2 display seen from 3 display heights
# R = fvvdp_display_geometry( (1920, 1080), diagonal_size_inches=47, distance_display_heights=3 )
# R.get_ppd()
#
# Some information about the effective FOV of VR headsets
# http://www.sitesinvr.com/viewer/htcvive/index.html
class fvvdp_display_geometry:

    def __init__(self, resolution, distance_m=None, distance_display_heights=None, fov_horizontal=None, fov_vertical=None, fov_diagonal=None, diagonal_size_inches=None) -> None:

        self.resolution = resolution
        
        ar = resolution[0]/resolution[1] # width/height
        
        self.fixed_ppd = None

        if not diagonal_size_inches is None:
            height_mm = math.sqrt( (diagonal_size_inches*25.4)**2 / (1+ar**2) )
            self.display_size_m = (ar*height_mm/1000, height_mm/1000)
                
        if (not distance_m is None) and (not distance_display_heights is None):
            raise RuntimeError( 'You can pass only one of: ''distance_m'', ''distance_display_heights''.' )
        
        if not distance_m is None:
            self.distance_m = distance_m;
        elif not distance_display_heights is None:
            if not hasattr( self, "display_size_m" ):
                raise RuntimeError( 'You need to specify display diagonal size ''diagonal_size_inches'' to specify viewing distance as ''distance_display_heights'' ' )
            self.distance_m = distance_display_heights * self.display_size_m[1]
        elif (not fov_horizontal is None) or (not fov_vertical is None) or (not fov_diagonal is None):
            # Default viewing distance for VR headsets
            self.distance_m = 3
        else:
            raise RuntimeError( 'Viewing distance must be specified as ''distance_m'' or ''distance_display_heights''.' )
        
        if ((not fov_horizontal is None) + (not fov_vertical is None) + (not fov_diagonal is None)) > 1:
            raise RuntimeError( 'You can pass only one of ''fov_horizontal'', ''fov_vertical'', ''fov_diagonal''. The other dimensions are inferred from the resolution assuming that the pixels are square.' )
        
        if not fov_horizontal is None:
            width_m = 2*math.tan( math.radians(fov_horizontal/2) )*self.distance_m
            self.display_size_m = (width_m, width_m/ar)
        elif not fov_vertical is None:
            height_m = 2*math.tan( math.radians(fov_vertical/2) )*self.distance_m
            self.display_size_m = (height_m*ar, height_m)
        elif not fov_diagonal is None:
            # Note that we cannot use Pythagorean theorem on degs -
            # we must operate on a distance measure
            # This is incorrect: height_deg = p.Results.fov_diagonal / sqrt( 1+ar^2 );
            
            distance_px = math.sqrt(self.resolution[0]**2 + self.resolution[1]**2) / (2.0 * math.tan( math.radians(fov_diagonal*0.5)) )
            height_deg = math.degrees(math.atan( self.resolution[1]/2 / distance_px ))*2
            
            height_m = 2*math.tan( math.radians(height_deg/2) )*self.distance_m
            self.display_size_m = (height_m*ar, height_m)
        
        self.display_size_deg = ( 2 * math.degrees(math.atan( self.display_size_m[0] / (2*self.distance_m) )), \
                                  2 * math.degrees(math.atan( self.display_size_m[1] / (2*self.distance_m) )) )
        
    # Get the number of pixels per degree
    #
    # ppd = R.get_ppd()
    # ppd = R.get_ppd(eccentricity)
    #
    # eccentricity is the viewing angle from the center in degrees. If
    # not specified, the central ppd value (for 0 eccentricity) is
    # returned.
    def get_ppd(self, eccentricity = None):
        
        # if ~isempty( dr.fixed_ppd )
        #     ppd = dr.fixed_ppd;
        #     return;
        # end
        
        # pixel size in the centre of the display
        pix_deg = 2*math.degrees(math.atan( 0.5*self.display_size_m[0]/self.resolution[0]/self.distance_m ))
        
        base_ppd = 1/pix_deg
        
        if eccentricity is None:
            return base_ppd
        else:
            delta = pix_deg/2
            tan_delta = math.tan(math.radians(delta))
            tan_a = torch.tan( torch.deg2rad(eccentricity) )
            
            ppd = base_ppd * (torch.tan(torch.deg2rad(eccentricity+delta))-tan_a)/tan_delta
            return ppd


    # Convert pixel positions into eccentricities for the given
    # display
    #
    # resolution_pix - image resolution as [width height] in pix
    # x_pix, y_pix - pixel coordinates generated with meshgrid,
    #   pixels indexed from 0
    # gaze_pix - [x y] of the gaze position, in pixels
    def pix2eccentricity( self, resolution_pix, x_pix, y_pix, gaze_pix ):
                        
        if not self.fixed_ppd is None:
            ecc = torch.sqrt( (x_pix-gaze_pix[0])**2 + (y_pix-gaze_pix[1])**2 )/self.fixed_ppd
        else:
            # Position the image in the centre
            shift_to_centre = -resolution_pix/2
            x_pix_rel = x_pix+shift_to_centre[0]
            y_pix_rel = y_pix+shift_to_centre[1]
            
            x_m = x_pix_rel * self.display_size_m[0] / self.resolution[0]
            y_m = y_pix_rel * self.display_size_m[1] / self.resolution[1]
            
            device = x_pix.device

            gaze_m = (gaze_pix + shift_to_centre) * torch.tensor(self.display_size_m) / torch.tensor(self.resolution)
            gaze_deg = torch.rad2deg(torch.atan( gaze_m/self.distance_m ))
            
            ecc = torch.sqrt( (torch.rad2deg(torch.atan(x_m/self.distance_m))-gaze_deg[0])**2 + (torch.rad2deg(torch.atan(y_m/self.distance_m))-gaze_deg[1])**2 )
        
        return ecc
        
    def get_resolution_magnification( self, eccentricity ):
            # Get the relative magnification of the resolution due to
            # eccentricity.
            # 
            # M = R.get_resolution_magnification(eccentricity)
            # 
            # eccentricity is the viewing angle from the center to the fixation point in degrees.
            
            if not self.fixed_ppd is None:
                M = torch( (1), device=eccentricity.device )
            else:            
                eccentricity = torch.minimum( eccentricity, torch.tensor((89.9)) ) # To avoid singulatities
                
                # pixel size in the centre of the display
                pix_rad = 2*math.atan( 0.5*self.display_size_m[0]/self.resolution[0]/self.distance_m )
                
                delta = pix_rad/2
                tan_delta = math.tan(delta)
                tan_a = torch.tan( torch.deg2rad(eccentricity) )
                
                M = (torch.tan(torch.deg2rad(eccentricity)+delta)-tan_a)/tan_delta

            return M

    def print(self):
        logging.info( 'Geometric display model:' )
        if hasattr( self, "fixed_ppd" ):
            logging.info( '  Fixed pixels-per-degree: {}'.format(self.fixed_ppd) )
        else:
            logging.info( '  Resolution: {w} x {h} pixels'.format( w=self.resolution[0], h=self.resolution[1] ) )
            logging.info( '  Display size: {w:.1f} x {h:.1f} cm'.format( w=self.display_size_m[0]*100, h=self.display_size_m[1]*100) )
            logging.info( '  Display size: {w:.2f} x {h:.2f} deg'.format( w=self.display_size_deg[0], h=self.display_size_deg[1] ) )
            logging.info( '  Viewing distance: {d:.3f} m'.format(d=self.distance_m) )
            logging.info( '  Pixels-per-degree (center): {ppd:.2f}'.format(ppd=self.get_ppd()) )

    @classmethod
    def load( cls, display_name ):
        models_file      = os.path.join(os.path.dirname(__file__), "fvvdp_data/display_models.json")
        models = json2dict(models_file)

        for mk in models:
            if mk == display_name:
                model = models[mk]
                assert "resolution" in model

                inches_to_meters = 0.0254

                W, H = model["resolution"]

                if "fov_diagonal" in model: fov_diagonal = model["fov_diagonal"]
                else:                       fov_diagonal = None

                if   "viewing_distance_meters" in model: distance_m = model["viewing_distance_meters"]
                elif "viewing_distance_inches" in model: distance_m = model["viewing_distance_inches"] * inches_to_meters
                else:                                    distance_m = None

                if   "diagonal_size_meters" in model: diag_size_inch = model["diagonal_size_meters"] / inches_to_meters
                elif "diagonal_size_inches" in model: diag_size_inch = model["diagonal_size_inches"] 
                else:                                 diag_size_inch = None

                obj = fvvdp_display_geometry( (W, H), distance_m=distance_m, fov_diagonal=fov_diagonal, diagonal_size_inches=diag_size_inch)
                return obj

        logging.error("Error: Display model '%s' not found in display_models.json" % display_name)
        return None




from setuptools import setup

with open("README.md", "r", encoding="utf8") as fh:
    long_description = fh.read()

setup(
    name='pyfvvdp',
    version='1.1.3',
    description='PyTorch code for \'FovVideoVDP\': a full-reference' \
                'visual quality metric that predicts the perceptual' \
                'difference between pairs of images or videos.',
    long_description=long_description,
    url='https://github.com/gfxdisp/FovVideoVDP',
    long_description_content_type='text/markdown',
    author='Rafał K. Mantiuk',
    author_email='mantiuk@gmail.com',
    license='Creative Commons Attribution-NonCommercial 4.0 International Public License',
    packages=['pyfvvdp', 'pyfvvdp/third_party'],
    package_data={'pyfvvdp': ['csf_cache/*.mat', 'fvvdp_data/*.json']},
    include_package_data=True,
    install_requires=['numpy>=1.17.3',
                      'scipy>=1.7.0',
                      'ffmpeg-python>=0.2.0',
                      'torch>=1.8.2',
                      'ffmpeg>=1.4',
                      'imageio>=2.19.5'],

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],

    entry_points={
        'console_scripts': [
            'fvvdp=pyfvvdp.run_fvvdp:main'
        ]
    }
)
